(* 
   file     : $CAT_DIR/ExtBin/ExtBinOracle.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : This file provides the ML function for the oracle declared
              in `ExtBin.thy'.
*)

signature ExtBinORACLE =
sig

  datatype CompArithOracleType = SimpData   of term
                               | SupInfData of term * term

  exception CompArithOracleExn of CompArithOracleType

  val computer_arithmetic_reasoner: Sign.sg * exn -> term

end;

structure ExtBinOracle: ExtBinORACLE =
struct

local

  open Basics;
  open IntSemantics;
  open IntSimplifier;
  open SupInf;

in

  datatype CompArithOracleType = SimpData   of term
                               | SupInfData of term * term;

  exception CompArithOracleExn of CompArithOracleType;


(* implementation of the oracle declared in ExtBin.thy; it calles the
   arithmetic simplifier or the SupInf decision procedure according
   to the type of the argument.
   type: Sign.sg * exn -> term  *)

  fun computer_arithmetic_reasoner
      (sign, CompArithOracleExn (SimpData x)) =
        (Const ("==>", PROP --> PROP --> PROP) $
        (IntTerm_to_term (int_simplify (term_to_IntTerm x))
         handle e => x) $ x)

    | computer_arithmetic_reasoner
      (sign, CompArithOracleExn (SupInfData (x, y))) =
        (Const ("==>", PROP --> PROP --> PROP) $
        (Const ("Trueprop", BOOL --> PROP) $ (SupInf x
         handle e => Const ("False", BOOL))) $ y)

    | computer_arithmetic_reasoner (_, _) = raise Error
        "Attempt to call ExtBin oracle with wrong arguments!!!";


end;  (* local *)

end;  (* struct *)


(* makes oracle globally accessible *)

open ExtBinOracle;



