(* 
   file     : $CAT_DIR/IntSimplifier/IntSemantics.ML
   platform : PolyML 3.2 and Isabelle94-8/HOL
   version  : 1.0b  (16.07.98)
   notes    : The language use by the Simplifier of CAT is defined.
*)

signature IntSEM =
sig

  datatype AnyTerm =
      Nat         of NatTerm
    | Int         of IntTerm

  and NatTerm =
      NatConstant of term * AnyTerm list
    | NatValue    of int
    | NatPlus     of NatTerm * NatTerm
    | NatMinus    of NatTerm * NatTerm
    | NatTimes    of NatTerm * NatTerm
    | NatDivide   of NatTerm * NatTerm
    | NatModulus  of NatTerm * NatTerm
    | NatPower    of NatTerm * NatTerm

  and IntTerm =
      IntNat      of NatTerm
    | IntConstant of term * AnyTerm list
    | IntValue    of int
    | IntAbs      of IntTerm
    | IntComp     of IntTerm
    | IntPlus     of IntTerm * IntTerm
    | IntMinus    of IntTerm * IntTerm
    | IntTimes    of IntTerm * IntTerm
    | IntDivide   of IntTerm * IntTerm
    | IntModulus  of IntTerm * IntTerm
    | IntPower    of IntTerm * IntTerm
    | IntAnd      of IntTerm * IntTerm
    | IntOr       of IntTerm * IntTerm
    | IntXOr      of IntTerm * IntTerm
    | IntNot      of IntTerm
    | IntASL      of IntTerm
    | IntASR      of IntTerm
    | IntRoL      of IntTerm
    | IntRoR      of IntTerm;

  val NatTerm_to_term: NatTerm -> term
  val IntTerm_to_term: IntTerm -> term
  val term_to_NatTerm: term    -> NatTerm
  val term_to_IntTerm: term    -> IntTerm

end;


structure IntSemantics: IntSEM =
struct

local

  open Basics;

in

(* signature of the CAT Simplifier language *)

  datatype AnyTerm =
      Nat         of NatTerm
    | Int         of IntTerm

  and NatTerm =
      NatConstant of term * AnyTerm list
    | NatValue    of int
    | NatPlus     of NatTerm * NatTerm
    | NatMinus    of NatTerm * NatTerm
    | NatTimes    of NatTerm * NatTerm
    | NatDivide   of NatTerm * NatTerm
    | NatModulus  of NatTerm * NatTerm
    | NatPower    of NatTerm * NatTerm

  and IntTerm =
      IntNat      of NatTerm
    | IntConstant of term * AnyTerm list
    | IntValue    of int
    | IntAbs      of IntTerm
    | IntComp     of IntTerm
    | IntPlus     of IntTerm * IntTerm
    | IntMinus    of IntTerm * IntTerm
    | IntTimes    of IntTerm * IntTerm
    | IntDivide   of IntTerm * IntTerm
    | IntModulus  of IntTerm * IntTerm
    | IntPower    of IntTerm * IntTerm
    | IntAnd      of IntTerm * IntTerm
    | IntOr       of IntTerm * IntTerm
    | IntXOr      of IntTerm * IntTerm
    | IntNot      of IntTerm
    | IntASL      of IntTerm
    | IntASR      of IntTerm
    | IntRoL      of IntTerm
    | IntRoR      of IntTerm;


(* auxiliary term *)

  val dummyTerm = Const ("dummy", INT);


(* composes an Isabelle constant/variable from the root symbol and a list
   of its arguments.
   type: term -> AnyTerm list -> term  *)

  fun mkConstTerm (Abs (a, T, _)) [Int u] =
        Abs (a, T, (IntTerm_to_term u))
    | mkConstTerm root [] =
        root
    | mkConstTerm root (x::xs) =
        let
          fun mkConst (Nat e) = NatTerm_to_term e
            | mkConst (Int e) = IntTerm_to_term e;
        in
          mkConstTerm (root $ (mkConst x)) xs
        end


(* embeds NatTerm into the set of Isabelle's nat terms.
   type: NatTerm -> term  *)

  and NatTerm_to_term nterm =
        case nterm of
            NatConstant (root, args) =>
              mkConstTerm root args
          | NatValue arg =>
	      nat_to_term arg
          | NatPlus (arg1, arg2) =>
	      Const ("op +", NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2
          | NatMinus (arg1, arg2) =>
	      Const ("op -",NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2
          | NatTimes (arg1, arg2) =>
	      Const ("op *",NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2
          | NatDivide (arg1, arg2) =>
	      Const ("div",NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2
          | NatModulus (arg1, arg2) =>
	      Const ("mod",NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2
          | NatPower (arg1, arg2) =>
	      Const ("op ^",NAT --> NAT --> NAT) $
	        NatTerm_to_term arg1 $
	        NatTerm_to_term arg2


(* embeds IntTerm into the set of Isabelle's int terms.
   type: IntTerm -> term  *)

  and IntTerm_to_term iterm =
        case iterm of
            IntNat arg =>
              Const ("znat", NAT --> INT) $ NatTerm_to_term arg
          | IntConstant (root, args) =>
              mkConstTerm root args
          | IntValue arg =>
	      int_to_term arg
          | IntAbs arg =>
              Const ("zmagnitude", INT --> INT) $
	        IntTerm_to_term arg
          | IntComp arg =>
              Const ("zminus", INT --> INT) $
	        IntTerm_to_term arg
          | IntPlus (arg1, arg2) =>
	      Const ("op +", INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntMinus (arg1, arg2) =>
	      Const ("op -",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntTimes (arg1, arg2) =>
	      Const ("op *",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntDivide (arg1, arg2) =>
	      Const ("op zdiv",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntModulus (arg1, arg2) =>
	      Const ("op zmod",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntPower (arg1, arg2) =>
	      Const ("op zexp",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntAnd (arg1, arg2) =>
	      Const ("op zand",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntOr (arg1, arg2) =>
	      Const ("op zor",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntXOr (arg1, arg2) =>
	      Const ("op zxor",INT --> INT --> INT) $
	        IntTerm_to_term arg1 $
	        IntTerm_to_term arg2
          | IntNot arg =>
              Const ("znot", INT --> INT) $
	        IntTerm_to_term arg
          | IntASL arg =>
              Const ("zasl", INT --> INT) $
	        IntTerm_to_term arg
          | IntASR arg =>
              Const ("zasr", INT --> INT) $
	        IntTerm_to_term arg
          | IntRoL arg =>
              Const ("zrol", INT --> INT) $
	        IntTerm_to_term arg
          | IntRoR arg =>
              Const ("zror", INT --> INT) $
	        IntTerm_to_term arg;


(* maps a list of Isabelle terms (argument list of a constant/variable)
   to a list of AnyTerms.
   type: (term -> AnyTerm) -> term list -> AnyTerm  *)

  fun mkConsts term_to_x args =
    let
      fun mkConst (Bound i) = term_to_x (Bound i)
        | mkConst arg =
            case (type_of arg handle e => Type ("",[])) of
                Type ("int",[]) =>
                  Int (term_to_IntTerm arg)
              | Type ("nat",[]) =>
                  Nat (term_to_NatTerm arg)
              | _ =>
                  term_to_x arg;
    in
      map mkConst args
    end


(* maps Isabelle nat terms to NatTerm.
   type: term -> NatTerm  *)

  and term_to_NatTerm nterm =
        case nterm of
            Const ("Suc", Type ("fun",
              [Type ("nat",[]), Type ("nat",[])])) $ arg =>
              NatValue (term_to_int nterm)
          | Const ("0", Type ("nat",[])) =>
              NatValue 0
	  | Const ("op +",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatPlus (term_to_NatTerm arg1, term_to_NatTerm arg2)
	  | Const ("op -",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatMinus (term_to_NatTerm arg1, term_to_NatTerm arg2)
	  | Const ("op *",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatTimes (term_to_NatTerm arg1, term_to_NatTerm arg2)
	  | Const ("div",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatDivide (term_to_NatTerm arg1, term_to_NatTerm arg2)
	  | Const ("mod",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatModulus (term_to_NatTerm arg1, term_to_NatTerm arg2)
	  | Const ("op ^",
              Type ("fun",
                [Type ("nat",[]), Type ("fun",
                  [Type ("nat",[]), Type ("nat",[])])])) $ arg1 $ arg2 =>
	      NatPower (term_to_NatTerm arg1, term_to_NatTerm arg2)
          | Bound _ =>
              NatConstant (nterm, [])
          | _ =>
              let
                val (root, args) = term_split nterm;
              in
                NatConstant (root, mkConsts (fn x => Nat (term_to_NatTerm x)) 
                                             args)
              end


(* maps Isabelle int terms to IntTerm.
   type: term -> IntTerm  *)

  and term_to_IntTerm iterm =
        case iterm of
	    Const ("znat",
              Type ("fun",
                [Type ("nat",[]), Type ("int",[])])) $ arg =>
	      IntNat (term_to_NatTerm arg)
          | Const ("PlusSign", Type ("int",[])) =>
              IntValue 0
          | Const ("MinusSign", Type ("int",[])) =>
              IntValue ~1
          | Const ("Bcons",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("bool",[]), Type ("int",[])])])) $ _ $ _ =>
	      IntValue (term_to_int iterm)
          | Const ("zsuc",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
	      IntPlus (term_to_IntTerm arg, IntValue 1)
	  | Const ("zpred",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
	      IntMinus (term_to_IntTerm arg, IntValue 1)
          | Const ("zmagnitude",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntAbs (term_to_IntTerm arg)
	  | Const ("zminus",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
	      IntComp (term_to_IntTerm arg)
	  | Const ("op +",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntPlus (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op -",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntMinus (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op *",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntTimes (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zdiv",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntDivide (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zmod",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntModulus (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zexp",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntPower (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zand",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntAnd (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zor",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntOr (term_to_IntTerm arg1, term_to_IntTerm arg2)
	  | Const ("op zxor",
              Type ("fun",
                [Type ("int",[]), Type ("fun",
                  [Type ("int",[]), Type ("int",[])])])) $ arg1 $ arg2 =>
	      IntXOr (term_to_IntTerm arg1, term_to_IntTerm arg2)
          | Const ("znot",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntNot (term_to_IntTerm arg)
          | Const ("zasl",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntASL (term_to_IntTerm arg)
          | Const ("zasr",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntASR (term_to_IntTerm arg)
          | Const ("zrol",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntRoL (term_to_IntTerm arg)
          | Const ("zror",
              Type ("fun",
                [Type ("int",[]), Type ("int",[])])) $ arg =>
              IntRoR (term_to_IntTerm arg)
          | Bound _ =>
              IntConstant (iterm, [])
          | Abs (a, T, u) =>
              IntConstant (Abs (a, T, dummyTerm),
                           mkConsts (fn x => Int (term_to_IntTerm x)) [u])
          | _ =>
              let
                val (root, args) = term_split iterm;
              in
                IntConstant (root,
                  mkConsts (fn x => Int (term_to_IntTerm x)) args)
              end;


end; (* local *)

end; (* struct *)
